import sys, os, socket, termcolor, logging
from paramiko import SSHClient
from .Lib import Lib
from ._get_info import _get_info
from ._exec_cmds import _exec_cmds

@Lib.add_methods_from(_get_info)
@Lib.add_methods_from(_exec_cmds)
class RasPyCheck():
    logger = None
    loglevel = None
    ip = None
    user = None
    password = None
    ssh_client = None
    keep_connection_open = None

    def __init__(self, ip, user, password, persist=False, debug=False):
        """
        Initializes an instance of RasPyCheck

        :param str ip: IP of the Raspberry Pi
        :param str user: Username to connect with
        :param str password: Password to connect with
        :param boolean persist: Keep connection alive (defaults to False)
        :param boolean debug: Gives more output (loglevel debug)
        """
        self.set_loglevel(debug)
        self.init_logger()
        self.ip = ip
        self.user = user
        self.password = password
        self.keep_connection_open = persist
        if self.keep_connection_open == True:
            self.setup_client()

    def set_loglevel(self, debug):
        """
        Sets loglevel to eiter DEBUG or ERROR
        """
        if debug == True:
            self.loglevel = logging.DEBUG
        else:
            self.loglevel = logging.ERROR

    def init_logger(self):
        """
        Initializes the logger
        """
        handler = logging.StreamHandler(sys.stdout)
        frm = logging.Formatter("[raspycheck] {asctime} - {levelname}: {message}", "%d.%m.%Y %H:%M%S", style="{")
        handler.setFormatter(frm)
        self.logger = logging.getLogger()
        self.logger.addHandler(handler)
        self.logger.setLevel(self.loglevel)

    def setup_client(self):
        """
        Creates a SSHClient instance and keeps it open until closed by calling
        self.close instance
        """
        self.ssh_client = SSHClient()
        self.ssh_client.load_system_host_keys()
        self.ssh_client.connect(self.ip, username=self.user, password=self.password)

    def close_instance_client(self):
        """
        Closes instance client
        """
        if self.ssh_client is not None:
            self.ssh_client.close()

    def cleanup_cmd_execution(self, stdin, stdout, stderr, client):
        """
        Cleanup after cmd execution. Closes streams and client if the latter is not persistant.

        :param paramiko.channel.ChannelStdinFile stdin: SSH stdin object
        :param paramiko.channel.ChannelFile stdout: SSH stdout object
        :param paramiko.channel.ChannelStderrFile stderr: SSH stderr object
        :param SSHClient client: Instance of SSHClient used for command execution, will only get close if not persisted
        """
        stdin.close()
        stdout.close()
        stderr.close()
        if self.ssh_client == None:
            client.close()

    def get_client(self):
        """
        Returns either the instance SSHClient or creates a new, temporary one
        """
        if self.ssh_client is not None:
            client = self.ssh_client
        else:
            client = SSHClient()
            client.load_system_host_keys()
            client.connect(self.ip, username=self.user, password=self.password)
        return client

    def run_command_get_output(self, cmd):
        """
        Executes command on raspberry pi and returns STDOUT

        :param str cmd: Command to execute on the Pi
        :returns: Response from the command
        """
        client = self.get_client()
        stdin, stdout, stderr = client.exec_command(cmd)
        response = stdout.read().decode("utf8")
        error = stderr.read().decode("utf8")
        self.logger.debug(response)
        if error is not None and error != "":
            self.logger.error(termcolor.colored(error, "red"))
        self.cleanup_cmd_execution(stdin, stdout, stderr, client)
        return response

    def run_command_get_return_code(self, cmd):
        """
        Executes command on Raspberry Pi and returns return code

        :param str cmd: Command to execute
        :returns: Return code from command execution
        """
        client = self.get_client()
        stdin, stdout, stderr = client.exec_command(cmd)
        return_code = stdout.channel.recv_exit_status()
        error = stderr.read().decode("utf8")
        self.logger.debug(return_code)
        if error is not None and error != "":
            self.logger.error(termcolor.colored(error, "red"))
        self.cleanup_cmd_execution(stdin, stdout, stderr, client)
        return return_code



def add_methods_from(*modules):
    def decorator(Class):
        for module in modules:
            for method in getattr(module, "__methods__"):
                setattr(Class, method.__name__, method)
        return Class
    return decorator

def register_method(methods):
    def register_method(method):
        methods.append(method)
        return method
    return register_method



from raspycheck.Lib import Lib

__methods__ = []
register_method = Lib.register_method(__methods__)

@register_method
def get_rpi_ip(self):
    """
    Returns IPv4 and IPv6 addresses from the Pi

    :returns: String containing IPv4 and IPv6 addresses, seperated by space
    """
    rpip = self.run_command_get_output('hostname -I').replace("\n", "")
    return rpip

@register_method
def get_rpi_hostname(self):
    """
    Returns hostname from the Pi

    :returns: String containing hostname of the Pi
    """
    hostname = self.run_command_get_output('hostname').replace("\n", "")
    return hostname

@register_method
def get_rpi_cpu_temp(self):
    """
    Returns CPU temperature of the pi

    :returns: String containing CPU temperature of the RPI
    """
    temp = self.run_command_get_output('vcgencmd measure_temp').replace("\n", "").replace("temp=", "")
    return temp

@register_method
def get_rpi_usb_devices(self):
    """
    Returns a list of the currently connected USB devices of the Pi

    :returns: List with USB devices
    """
    devices = self.run_command_get_output('lsusb')
    devices = devices.split("\n")
    devices = list(filter(None, devices))
    return devices

@register_method
def get_rpi_version(self):
    """
    Returns version of RPI

    :returns: String with version information
    """
    version = self.run_command_get_output('cat /proc/version').replace("\n", "")
    return version

@register_method
def get_rpi_free_memory(self):
    """
    Returns info about how much free memory is available

    :returns: Free memory
    """
    memory = self.run_command_get_output('free').split()[9]
    return memory

@register_method
def get_rpi_total_memory(self):
    """
    Returns info about how much total memory is available

    :returns: Total memory
    """
    memory = self.run_command_get_output('free').split()[7]
    return memory

@register_method
def get_rpi_available_essids(self):
    """
    Returns list of ESSIDs available to the Pi

    :returns: List of ESSIDs
    """
    essids = self.run_command_get_output('iwlist wlan0 scan | grep ESSID')
    return essids

@register_method
def get_rpi_date(self):
    """
    Returns date and time as used by the Pi

    :returns: Date of the RPI
    """
    date = self.run_command_get_output('date').replace("\n", "")
    return date

@register_method
def get_rpi_list_installed_packages(self):
    """
    Returns list with installed packages
    """
    packages = self.run_command_get_output('dpkg --get-selections')
    return packages

@register_method
def get_all_rpi_info(self):
    """
    Returns dictionary with all available RPI information
    """
    return {
        'ip': self.get_rpi_ip(),
        'hostname': self.get_rpi_hostname(),
        'cpu_temp': self.get_rpi_cpu_temp(),
        'usb_devices': self.get_rpi_usb_devices(),
        'version': self.get_rpi_version(),
        'free_memory': self.get_rpi_free_memory(),
        'total_memory': self.get_rpi_total_memory(),
        'date': self.get_rpi_date()
    }



from raspycheck.Lib import Lib

__methods__ = []
register_method = Lib.register_method(__methods__)

@register_method
def cmd_reboot_pi(self):
    """
    Reboots RPI

    :returns: Return code of reboot cmd
    """
    ret = self.run_command_get_return_code('reboot')
    return ret



from setuptools import setup, find_packages

with open('README.md') as readme_file:
    README = readme_file.read()

with open('HISTORY.md') as history_file:
    HISTORY = history_file.read()

setup_args = dict(
    name='raspycheck',
    version='0.0.2',
    description='Useful tool to execute commands via SSH on a RPI or grab system info from said RPI',
    long_description_content_type="text/markdown",
    long_description=README + '\n\n' + HISTORY,
    license='MIT',
    packages=find_packages(),
    author='Felix Harenbrock',
    author_email='felix.harenbrock@gmx.de',
    keywords=['Raspberry Pi', 'Raspi', 'RPI'],
    url='https://github.com/kampfhamster309/raspycheck',
    test_suite='nose.collector',
    tests_require=['nose']
    #download_url='https://pypi.org/project/elastictools/'
)

install_requires = [
    'paramiko==2.7.2',
    'termcolor==1.1.0'
]

if __name__ == '__main__':
    setup(**setup_args, install_requires=install_requires)
